## two dynamic variables
handler.stack <- NULL  ## place holder for .First.lib
handler.mask <- NULL   ## place holder for .First.lib
toplevel.established <- NULL # place holder for .First.lib
catch<-function(tag, expr) {
    "__CATCH_TAG__" <- as.character(tag)
    "__CATCH_THROWER__"<- make.thrower(return(get("__CATCH_VALUE__")))
    "__CATCH_VALUE__" <- NULL
    expr
}

make.thrower <- function(expr) function() expr
throw<-function(tag, expr, no.tag = tag.not.found) {
    value <- expr # forces evaluation of expr
    tag <- as.character(tag)
    env <- get.target.frame(tag, "__CATCH_TAG__")
    assign("__CATCH_VALUE__", value, env = env)
    fun <- get("__CATCH_THROWER__", env = env)
    fun()
}

##****parent.env with arg > 1 doesn't seem to work as advertised??
get.target.frame <- function(tag, name) {
    n <- sys.nframe()
    if (n > 1)
        for (i in (n-1):1) {
            env <- sys.frame(i)
            if (exists(name, env = env) &&
                get(name, env = env) == tag)
                return(env)
        }
    stop(paste("no catch for tag \"", tag, "\"", sep = ""))
} 
add.to.handler.stack <- function(handler, class, exiting, stack) {
    if (is.null(stack)) index <- 1
    else index <- stack$index + 1
    list(handler = handler, class = class, exiting = exiting,
         next.handler = stack, index = index)
}
add.to.handler.mask <- function(h, mask)
    c(h$index, mask)
is.handler.masked <- function(h, mask)
    if (match(h$index, mask, 0)) TRUE else FALSE
call.handler <- function(h, c) {
    if (h$exiting)
        throw("__TRY_CATCH__", list(throw = TRUE, handler = h, condition = c))
    else {
        mask <- add.to.handler.mask(h, handler.mask())
        dynamic.bind(h$handler(c), handler.mask = mask)
    }
}
find.handler <- function(h, c) {
    mask <- handler.mask()
    h <- handler.stack()
    while (! is.null(h))
        if (! is.handler.masked(h, mask) && inherits(c, h$class))
            return(h)
        else h <- h$next.handler
    raise.exception(no.condition.handler.exception(c))
}
handler.index <- function(h) if (is.null(h)) 0 else h$index
signal.condition <- function(c) {
    check.toplevel.context()
    if (! inherits(c, "condition"))
        c <- simple.condition(c)
    call.handler(find.handler(h, c), c)
}
my.stop <- function(e, call. = TRUE) {
    if (! inherits(e, "exception"))
        e <- simple.exception(e, if (call.) sys.call(1) else NULL)
    signal.condition(e)
    abort()
}
raise.exception <- function(e) my.stop(e)
my.warning <- function(w) {
    if (! inherits(w, "warning"))
       w <- simple.warning(w, sys.call(1))
    signal.condition(w)
}
abort <- function(call. = FALSE) {
    repeat {
        if (call.)
            signal.condition(abort.restart(sys.call(1)))
        else 
            signal.condition(the.abort.restart)
    }
}
with.handlers <- function(expr, ...) {
    check.toplevel.context()
    stack <- handler.stack()
    handlers <- rev(list(...))
    classes <- names(handlers)
    for (i in seq(along = handlers))
        stack <- add.to.handler.stack(handlers[[i]], classes[i], FALSE, stack)
    dynamic.bind(expr, handler.stack = stack)
}
muffle.warnings <- function(expr)
    with.handlers(expr, warning = function(w) {})
try.catch <- function(expr, ..., finally = NULL) {
    check.toplevel.context()
    on.exit(finally)
    handlers <- rev(list(...))
    stack <- handler.stack()
    base <- handler.index(stack)
    classes <- names(handlers)
    for (i in seq(along = handlers))
        stack <- add.to.handler.stack(handlers[[i]], classes[i], TRUE, stack)
    result <- catch("__TRY_CATCH__",
                    dynamic.bind(list(throw = FALSE, value = expr),
                                 handler.stack = stack))
    if (result$throw) {
        idx <- handler.index(result$handler)
        if (idx > base) {
            h <- result$handler$handler
            c <- result$condition
            result <- catch("__TRY_CATCH__", list(throw = FALSE, value = h(c)))
        }
    }
    if (result$throw) {
        on.exit()
        finally
        throw("__TRY_CATCH__", result)
    }
    else result$value
}
print.condition <- function(c, ...) {
    msg <- condition.message(c)
    call <- condition.call(c)
    class <- class(c)[1]
    if (! is.null(call))
        cat("<", class, " in ", deparse(call), ": ", msg, ">\n", sep="")
    else
        cat("<", class, ": ", msg, ">\n", sep="")
}

condition.message <- function(c) UseMethod("condition.message", c)
condition.call <- function(c) UseMethod("condition.call", c)

condition.message.condition <- function(c) c$message
condition.call.condition <- function(c) c$call
simple.condition <- function(message, call = NULL) {
    class <- c("simple.condition", "condition")
    structure(list(message=as.character(message), call = call), class=class)
}

simple.exception <- function(message, call = NULL) {
    class <- c("simple.exception", "exception", "condition")
    structure(list(message=as.character(message), call = call), class=class)
}

simple.warning <- function(message, call = NULL) {
    class <- c("simple.warning", "warning", "condition")
    structure(list(message=as.character(message), call = call), class=class)
}
abort.restart <- function(call = NULL)
    structure(list(message="abort", call = call),
              class = c("abort", "restart", "condition"))
the.abort.restart <- abort.restart()
no.condition.handler.exception <- function(c)
    structure(list(message = paste("no condition handler for", class(c)[1]),
                   condition = c),
              class = c("no.condition.handler.exception",
                        "exception", "condition"))
default.handler <- function(e) {
    UseMethod("default.handler", e)
}

default.handler.warning <- function(w) {
    cat("Warning:", w$message, "\n")
}

default.handler.exception <- function(e) {
    prerr <- function(...) cat(..., "\n", sep = "") # to error output??
    class <- class(e)[1]
    call <- condition.call(e)
    msg <- condition.message(e)
    if (! is.null(call))
        prerr("unhandled exception (", class, ") in ", call, ": ", msg)
    else
        prerr("unhandled exception (", class, "): ", msg)
    abort()
}
error.converter <- function() {
    .Call("ResetErrorHandling")
    raise.exception(simple.exception(geterrmessage()))
}

check.toplevel.context <- function()
    if (! toplevel.established()) stop("suicide: no toplevel context")
toplevel<- function(expr) {
    error <- getOption("error")
    show <- getOption("show.error.messages")
    if (is.null(show)) show <- TRUE
    op <- list(error = error, show.error.messages = show)
    on.exit(options(op))
    options(error = error.converter, show.error.messages = FALSE)
    failsafe.handler <- function(c) list(condition = c)
    dynamic.bind(try.catch(with.handlers(list(value = expr),
                                         exception = default.handler,
                                         warning = default.handler),
                           exception = failsafe.handler,
                           abort = failsafe.handler),
                 toplevel.established = TRUE,
                 handler.stack = NULL)
}
Quit <- function(save = "default", status = 0, runLast = TRUE) {
    cond <- structure(list(save = save, status = status, runLast = runLast),
                      class = c("quit", "restart", "condition"))
    signal.condition(cond)
}

Repl <- function() {
    do.one <- function() {
        expr <- parse(prompt = "===> ")
        print(eval(expr, .GlobalEnv))
     }
     try.catch(while (TRUE)
                   try.catch(do.one(), abort = function(c) {}),
               quit = function(c) c)
}

mainloop <- function() {
    while (TRUE) {
        result <- toplevel(Repl())
        if (is.null(result$value)) {
            prerr <- function(...) cat(..., "\n", sep = "") # to error output??
            cond <- result$condition
            class <- class(cond)[1]
            call <- condition.call(cond)
            msg <- condition.message(cond)
            if (! is.null(call))
                prerr("unhandled condition (", class, ") in ", call, ": ", msg)
            else
                prerr("unhandled condition (", class, "): ", msg)
        }
        else return(unclass(result$value))
    }
}
.First.lib <- function(lib, pkg) {
    library.dynam(pkg, pkg, lib)

    require(dynvars)
    handler.stack <<- dynamic.variable()
    handler.mask <<- dynamic.variable()
    toplevel.established <<- dynamic.variable(FALSE)
}
